/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::ArMemMemoryViewerWidgetController
 * \author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * \date       2020
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ArMemMemoryViewerWidgetController.h"

#include <string>

#include <QTimer>

#include <ArmarXCore/observers/variant/Variant.h>


namespace armarx
{
    QString ArMemMemoryViewerWidgetController::GetWidgetName()
    {
        return "MemoryX.ArMemMemoryViewer";
    }


    ArMemMemoryViewerWidgetController::ArMemMemoryViewerWidgetController()
    {
        widget.setupUi(getWidget());

        // Update timer
        updateTimer = new QTimer(this);
        updateTimer->setInterval(static_cast<int>(std::round(1000 / widget.updateFrequencySpinBox->value())));

        using This = ArMemMemoryViewerWidgetController;

        connect(this, &This::connected, this, &This::update);
        connect(widget.updateButton, &QPushButton::pressed, this, &This::update);
        connect(updateTimer, &QTimer::timeout, this, &This::update);

        widget.updateFrequencySpinBox->setEnabled(widget.updateAutoCheckBox->isChecked());
        connect(widget.updateAutoCheckBox, &QCheckBox::toggled, widget.updateFrequencySpinBox, &QDoubleSpinBox::setEnabled);

        connect(widget.updateFrequencySpinBox, &QDoubleSpinBox::editingFinished, [this]()
        {
            updateTimer->setInterval(static_cast<int>(std::round(1000 / widget.updateFrequencySpinBox->value())));
        });
        connect(widget.updateAutoCheckBox, &QCheckBox::toggled, this, [this](bool checked)
        {
            if (checked)
            {
                updateTimer->start();
            }
            else
            {
                updateTimer->stop();
            }
        });

        queryGui.setWidget(widget.queryTabWidget);
    }

    ArMemMemoryViewerWidgetController::~ArMemMemoryViewerWidgetController()
    {
    }


    void ArMemMemoryViewerWidgetController::onInitComponent()
    {
        if (memoryName.size() > 0)
        {
            usingProxy(memoryName);
        }
        if (debugObserverName.size() > 0)
        {
            usingProxy(debugObserverName);
        }

        emit inited();
    }

    void ArMemMemoryViewerWidgetController::onConnectComponent()
    {
        if (memoryName.size() > 0)
        {
            getProxy(memory, memoryName);
            memoryReader = armem::MemoryReader(memory);
        }
        // DebugObserver is optional (check for null on every call)
        if (!debugObserverName.empty())
        {
            getProxy(debugObserver, debugObserverName, false, "", false);
        }

        if (widget.updateAutoCheckBox->isChecked())
        {
            updateTimer->stop();
        }

        emit connected();
    }


    void ArMemMemoryViewerWidgetController::onDisconnectComponent()
    {
        if (updateTimer)
        {
            updateTimer->stop();
        }
    }


    void ArMemMemoryViewerWidgetController::update()
    {
        if (!memoryReader)
        {
            return;
        }

        armem::Time start = armem::Time::now();
        armem::MemoryPtr queryResult;
        {
            armem::query::EntityQueryPtr entityQuery = queryGui.entityQuery();
            ARMARX_CHECK(entityQuery);
            armem::query::ProviderSegmentQueryPtr providerQuery = new armem::query::provider::All({entityQuery});
            armem::query::CoreSegmentQueryPtr coreQuery = new armem::query::core::All({providerQuery});
            armem::query::MemoryQueryPtr memoryQuery = new armem::query::memory::All({coreQuery});

            queryResult = memoryReader.queryToCpp(memoryQuery);
        }
        ARMARX_VERBOSE << "Memory query took " << (armem::Time::now() - start).toMilliSecondsDouble() << " ms.";

        if (debugObserver)
        {
            debugObserver->setDebugDatafield(debugObserverChannelName, "Memory Query [ms]", new Variant((armem::Time::now() - start).toMilliSecondsDouble()));
        }

        if (queryResult)
        {
            updateMemoryTree(*queryResult, *widget.memoryTree);
        }
        else
        {
            ARMARX_ERROR << "No query result";
        }
    }

    void ArMemMemoryViewerWidgetController::updateMemoryTree(armem::Memory& memory, QTreeWidget& tree)
    {
        armem::Time start = armem::Time::now();

        builder.updateChildren(memory, &tree);

        ARMARX_VERBOSE << "GUI update took " << (armem::Time::now() - start).toMilliSecondsDouble() << " ms.";
        if (debugObserver)
        {
            debugObserver->setDebugDatafield(debugObserverChannelName, "GUI Update [ms]", new Variant((armem::Time::now() - start).toMilliSecondsDouble()));
        }
    }


    const static std::string CONFIG_KEY_MEMORY = "MemoryName";
    const static std::string CONFIG_KEY_DEBUG_OBSERVER = "DebugObserverName";

    void ArMemMemoryViewerWidgetController::loadSettings(QSettings* settings)
    {
        memoryName = settings->value(QString::fromStdString(CONFIG_KEY_MEMORY),
                                     "Memory").toString().toStdString();
        debugObserverName = settings->value(QString::fromStdString(CONFIG_KEY_DEBUG_OBSERVER),
                                            "DebugObserver").toString().toStdString();
    }
    void ArMemMemoryViewerWidgetController::saveSettings(QSettings* settings)
    {
        settings->setValue(QString::fromStdString(CONFIG_KEY_MEMORY),
                           QString::fromStdString(memoryName));
        settings->setValue(QString::fromStdString(CONFIG_KEY_DEBUG_OBSERVER),
                           QString::fromStdString(debugObserverName));
    }

    QPointer<QDialog> ArMemMemoryViewerWidgetController::getConfigDialog(QWidget* parent)
    {
        if (!configDialog)
        {
            configDialog = new SimpleConfigDialog(parent);
            configDialog->addProxyFinder<armem::MemoryInterfacePrx>({CONFIG_KEY_MEMORY, "Memory", "*Memory"});
            configDialog->addProxyFinder<armarx::DebugObserverInterfacePrx>({CONFIG_KEY_DEBUG_OBSERVER, "Debug Observer", "DebugObserver"});

        }
        return qobject_cast<QDialog*>(configDialog);
    }

    void ArMemMemoryViewerWidgetController::configured()
    {
        if (configDialog)
        {
            memoryName = configDialog->getProxyName(CONFIG_KEY_MEMORY);
            debugObserverName = configDialog->getProxyName(CONFIG_KEY_DEBUG_OBSERVER);
        }
    }
}
