#include "MemoryQueryWidget.h"

#include <QCheckBox>
#include <QComboBox>
#include <QDateTimeEdit>
#include <QFormLayout>
#include <QGridLayout>
#include <QLabel>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QSpinBox>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/armem/core/Time.h>
#include <RobotAPI/libraries/armem/core/ice_conversions.h>


namespace armarx::armem
{

    MemoryQueryWidget::MemoryQueryWidget()
    {
        connect(this, &This::widgetSet, this, &This::initWidget);
    }

    void MemoryQueryWidget::setWidget(QTabWidget* tabWidget)
    {
        this->widget = tabWidget;

        emit widgetSet();
    }

    armem::query::EntityQueryPtr MemoryQueryWidget::entityQuery()
    {
        ARMARX_CHECK_NOT_NULL(entityWidget);
        return entityWidget->query;
    }

    void MemoryQueryWidget::initWidget()
    {
        if (!widget)
        {
            return;
        }
        widget->clear();

        entityWidget = std::make_unique<EntityQueryWidget>();
        widget->addTab(entityWidget->widget, QString("Entity"));
    }



    EntityQueryWidget::EntityQueryWidget()
    {
        widget = new QWidget();
        QVBoxLayout* pageLayout = new QVBoxLayout();
        widget->setLayout(pageLayout);

        connect(this, &This::queryOutdated, this, &This::updateQuery);

        {
            QHBoxLayout* typeLayout = new QHBoxLayout();

            queryComboBox = new QComboBox();
            queryComboBox->addItems({"All", "Single", "Time Range", "Index Range"});

            dataCheckBox = new QCheckBox("Get Data");
            dataCheckBox->setChecked(true);

            typeLayout->addWidget(queryComboBox);
            typeLayout->addWidget(dataCheckBox);

            connect(queryComboBox, &QComboBox::currentTextChanged, this, &This::updateVisibleForm);

            connect(queryComboBox, &QComboBox::currentTextChanged, this, &This::queryOutdated);
            connect(dataCheckBox, &QCheckBox::stateChanged, this, &This::queryOutdated);

            pageLayout->addLayout(typeLayout);
        }

        queryForms.clear();
        {
            auto r = queryForms.emplace("All", std::make_unique<EntityFormAll>());
            pageLayout->addWidget(r.first->second->widget);
            connect(r.first->second.get(), &EntityForm::queryChanged, this, &This::updateQuery);
        }
        {
            auto r = queryForms.emplace("Single", std::make_unique<EntityFormSingle>());
            pageLayout->addWidget(r.first->second->widget);
            connect(r.first->second.get(), &EntityForm::queryChanged, this, &This::updateQuery);
        }
        {
            auto r = queryForms.emplace("Time Range", std::make_unique<EntityFormTimeRange>());
            pageLayout->addWidget(r.first->second->widget);
            connect(r.first->second.get(), &EntityForm::queryChanged, this, &This::updateQuery);
        }
        {
            auto r = queryForms.emplace("Index Range", std::make_unique<EntityFormIndexRange>());
            pageLayout->addWidget(r.first->second->widget);
            connect(r.first->second.get(), &EntityForm::queryChanged, this, &This::updateQuery);
        }

        updateVisibleForm(queryComboBox->currentText());
        updateQuery();
    }

    void EntityQueryWidget::updateQuery()
    {
        this->query = makeQuery();
        emit queryChanged();
    }

    void EntityQueryWidget::updateVisibleForm(QString visible)
    {
        hideAllForms();
        auto it = queryForms.find(visible);
        ARMARX_CHECK(it != queryForms.end());
        it->second->widget->setVisible(true);
    }

    armem::query::EntityQueryPtr EntityQueryWidget::makeQuery()
    {
        armem::query::EntityQueryPtr query = queryForms.at(queryComboBox->currentText())->makeQuery();
        ARMARX_CHECK_NOT_NULL(query);

        query->withData = dataCheckBox->isChecked();

        return query;
    }

    void EntityQueryWidget::hideAllForms()
    {
        for (auto& [_, form] : queryForms)
        {
            form->widget->setVisible(false);
        }
    }


    EntityForm::EntityForm()
    {
        widget = new QWidget();
    }

    void EntityForm::setDisplayFormat(QDateTimeEdit* dt)
    {
        dt->setDisplayFormat("yyyy-MM-dd HH:mm:ss");
    }


    EntityFormSingle::EntityFormSingle()
    {
        const QDateTime now = QDateTime::currentDateTime();

        label = new QLabel("Timestamp:");

        dateTime = new QDateTimeEdit(now);
        dateTime->setDisabled(true);
        setDisplayFormat(dateTime);

        latest = new QCheckBox("Latest");
        latest->setChecked(true);

        QGridLayout* layout = new QGridLayout(widget);
        layout->addWidget(label, 0, 0);
        layout->addWidget(dateTime, 0, 1);
        layout->addWidget(latest, 0, 2);

        connect(latest, &QCheckBox::toggled, dateTime, &QDateTimeEdit::setDisabled);

        connect(dateTime, &QDateTimeEdit::dateTimeChanged, this, &EntityForm::queryChanged);
        connect(latest, &QCheckBox::toggled, this, &EntityForm::queryChanged);
    }

    query::EntityQueryPtr EntityFormSingle::makeQuery()
    {
        query::entity::SinglePtr query = new armem::query::entity::Single();

        if (!latest->isChecked())
        {
            Time time = Time::milliSeconds(dateTime->dateTime().toMSecsSinceEpoch());
            toIce(query->timestamp, time);
        }

        return query;
    }


    EntityFormTimeRange::EntityFormTimeRange()
    {
        const QDateTime now = QDateTime::currentDateTime();

        fromLabel = new QLabel("From:");

        fromDateTime = new QDateTimeEdit(now.addSecs(-60));
        fromDateTime->setDisabled(true);
        setDisplayFormat(fromDateTime);

        fromBegin = new QCheckBox("Begin");
        fromBegin->setChecked(true);

        toLabel = new QLabel("To:");

        toDateTime = new QDateTimeEdit(now.addSecs(60));
        toDateTime->setDisabled(true);
        setDisplayFormat(toDateTime);

        toEnd = new QCheckBox("End");
        toEnd->setChecked(true);

        QGridLayout* layout = new QGridLayout(widget);
        int row = 0;

        layout->addWidget(fromLabel, row, 0);
        layout->addWidget(fromDateTime, row, 1);
        layout->addWidget(fromBegin, row, 2);
        ++row;

        layout->addWidget(toLabel, row, 0);
        layout->addWidget(toDateTime, row, 1);
        layout->addWidget(toEnd, row, 2);
        ++row;

        connect(fromBegin, &QCheckBox::toggled, fromDateTime, &QDateTimeEdit::setDisabled);
        connect(toEnd, &QCheckBox::toggled, toDateTime, &QDateTimeEdit::setDisabled);

        connect(fromBegin, &QCheckBox::toggled, this, &EntityForm::queryChanged);
        connect(fromDateTime, &QDateTimeEdit::dateTimeChanged, this, &EntityForm::queryChanged);
        connect(toEnd, &QCheckBox::toggled, this, &EntityForm::queryChanged);
        connect(toDateTime, &QDateTimeEdit::dateTimeChanged, this, &EntityForm::queryChanged);
    }

    query::EntityQueryPtr EntityFormTimeRange::makeQuery()
    {
        query::entity::TimeRangePtr query = new armem::query::entity::TimeRange();

        if (!fromBegin->isChecked())
        {
            Time time = Time::milliSeconds(fromDateTime->dateTime().toMSecsSinceEpoch());
            toIce(query->minTimestamp, time);
        }
        if (!toEnd->isChecked())
        {
            Time time = Time::milliSeconds(toDateTime->dateTime().toMSecsSinceEpoch());
            toIce(query->maxTimestamp, time);
        }

        return query;
    }


    EntityFormIndexRange::EntityFormIndexRange()
    {
        firstLabel = new QLabel("First:");

        firstSpinBox = new QSpinBox();
        firstSpinBox->setDisabled(true);

        firstBegin = new QCheckBox("Begin");
        firstBegin->setChecked(true);

        lastLabel = new QLabel("To:");

        lastSpinBox = new QSpinBox();
        lastSpinBox->setDisabled(true);

        lastEnd = new QCheckBox("End");
        lastEnd->setChecked(true);

        std::initializer_list<QSpinBox*> sbs { firstSpinBox, lastSpinBox };
        for (QSpinBox* sb : sbs)
        {
            sb->setMinimum(-1000);
            sb->setMaximum(1000);
        }
        firstSpinBox->setValue(0);
        lastSpinBox->setValue(-1);


        QGridLayout* layout = new QGridLayout(widget);
        int row = 0;

        layout->addWidget(firstLabel, row, 0);
        layout->addWidget(firstSpinBox, row, 1);
        layout->addWidget(firstBegin, row, 2);
        ++row;

        layout->addWidget(lastLabel, row, 0);
        layout->addWidget(lastSpinBox, row, 1);
        layout->addWidget(lastEnd, row, 2);
        ++row;

        connect(firstBegin, &QCheckBox::toggled, firstSpinBox, &QSpinBox::setDisabled);
        connect(lastEnd, &QCheckBox::toggled, lastSpinBox, &QSpinBox::setDisabled);

        connect(firstBegin, &QCheckBox::toggled, this, &EntityForm::queryChanged);
        connect(firstSpinBox, QOverload<int>::of(&QSpinBox::valueChanged), this, &EntityForm::queryChanged);
        connect(lastEnd, &QCheckBox::toggled, this, &EntityForm::queryChanged);
        connect(lastSpinBox, QOverload<int>::of(&QSpinBox::valueChanged), this, &EntityForm::queryChanged);
    }

    query::EntityQueryPtr EntityFormIndexRange::makeQuery()
    {
        query::entity::IndexRangePtr query = new armem::query::entity::IndexRange();

        if (!firstBegin->isChecked())
        {
            query->first = firstSpinBox->value();
        }
        if (!lastEnd->isChecked())
        {
            query->last = lastSpinBox->value();
        }

        return query;
    }

    query::EntityQueryPtr EntityFormAll::makeQuery()
    {
        return new armem::query::entity::All();
    }

}
