#include "MemoryTreeWidgetBuilder.h"

#include <SimoxUtility/algorithm/string.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/gui-plugins/ObjectPoseGui/TreeWidgetBuilder.h>

#include <RobotAPI/libraries/armem/component/QTreeWidgetAronDataVisitor.h>


namespace armarx::armem
{

    MemoryTreeWidgetBuilder::MemoryTreeWidgetBuilder()
    {
        memoryBuilder.setMakeItemFn([this](const std::string & name, const Memory * memory)
        {
            return makeItem(name, memory);
        });
        memoryBuilder.setUpdateItemFn([this](const std::string&, const Memory * memory, QTreeWidgetItem * memoryItem)
        {
            updateItem(memory, memoryItem);
            if (memoryItem)
            {
                updateChildren(*memory, memoryItem);
            }
            return true;
        });


        coreSegmentBuilder.setMakeItemFn([this](const std::string & name, const CoreSegmentPtr & coreSeg)
        {
            return makeItem(name, coreSeg.get());
        });
        coreSegmentBuilder.setUpdateItemFn([this](const std::string&, const CoreSegmentPtr & coreSeg, QTreeWidgetItem * coreSegItem)
        {
            updateItem(coreSeg.get(), coreSegItem);
            if (coreSeg)
            {
                updateChildren(*coreSeg, coreSegItem);
            }
            return true;
        });

        provSegmentBuilder.setMakeItemFn([this](const std::string & name, const ProviderSegmentPtr & provSeg)
        {
            return makeItem(name, provSeg.get());
        });
        provSegmentBuilder.setUpdateItemFn([this](const std::string&, const ProviderSegmentPtr & provSeg, QTreeWidgetItem * provSegItem)
        {
            updateItem(provSeg.get(), provSegItem);
            if (provSeg)
            {
                updateChildren(*provSeg, provSegItem);
            }
            return true;
        });

        entityBuilder.setMakeItemFn([this](const std::string & name, const EntityPtr & entity)
        {
            return makeItem(name, entity.get());
        });
        entityBuilder.setUpdateItemFn([this](const std::string&, const EntityPtr & entity, QTreeWidgetItem * entityItem)
        {
            updateItem(entity.get(), entityItem);
            if (entity)
            {
                updateChildren(*entity, entityItem);
            }
            return true;
        });

        snapshotBuilder.setMakeItemFn([this](const armem::Time & time, const EntitySnapshotPtr & snapshot)
        {
            QTreeWidgetItem* item = makeItem(toDateTimeMilliSeconds(time), snapshot.get());
            item->setData(0, Qt::ItemDataRole::UserRole, QVariant(static_cast<qlonglong>(snapshot->time().toMicroSeconds())));
            return item;
        });
        snapshotBuilder.setCompareFn([](const auto & pair, QTreeWidgetItem * item)
        {
            const EntitySnapshotPtr& snapshot = pair.second;
            return armarx::detail::compare(static_cast<qlonglong>(snapshot->time().toMicroSeconds()),
                                           item->data(0, Qt::ItemDataRole::UserRole).toLongLong());
        });
        snapshotBuilder.setUpdateItemFn([this](const armem::Time&, const EntitySnapshotPtr & snapshot, QTreeWidgetItem * snapshotItem)
        {
            updateItem(snapshot.get(), snapshotItem);
            if (snapshot)
            {
                updateChildren(*snapshot, snapshotItem);
            }
            else
            {
                while (snapshotItem->childCount() > 0)
                {
                    delete snapshotItem->takeChild(snapshotItem->childCount() - 1);
                }
            }
            return true;
        });

        instanceBuilder.setMakeItemFn([this](const EntityDataPtr & instance)
        {
            return makeItem("", instance.get());
        });
        instanceBuilder.setCompareFn([](const EntityDataPtr & lhs, QTreeWidgetItem * rhsItem)
        {
            return armarx::detail::compare(lhs->index, rhsItem->text(0).toInt());
        });
        instanceBuilder.setUpdateItemFn([this](const EntityDataPtr & instance, QTreeWidgetItem * instanceItem)
        {
            updateItem(instance.get(), instanceItem);
            if (instance)
            {
                updateChildren(*instance, instanceItem);
            }
            return true;
        });

    }

    void MemoryTreeWidgetBuilder::updateChildren(const armem::Memory& memory, QTreeWidget* tree)
    {
        updateChildren(std::map<string, const armem::Memory*> {{memory.name(), &memory}}, tree);
    }

    void MemoryTreeWidgetBuilder::updateChildren(const std::map<string, const armem::Memory*>& memories, QTreeWidget* tree)
    {
        memoryBuilder.updateTree(tree, memories);
    }


    void MemoryTreeWidgetBuilder::updateChildren(const armem::Memory& memory, QTreeWidgetItem* memoryItem)
    {
        coreSegmentBuilder.updateTree(memoryItem, memory.coreSegments);
    }

    void MemoryTreeWidgetBuilder::updateChildren(const armem::CoreSegment& coreSeg, QTreeWidgetItem* coreSegItem)
    {
        provSegmentBuilder.updateTree(coreSegItem, coreSeg.providerSegments);
    }

    void MemoryTreeWidgetBuilder::updateChildren(const armem::ProviderSegment& provSeg, QTreeWidgetItem* provSegItem)
    {
        entityBuilder.updateTree(provSegItem, provSeg.entities);
    }

    void MemoryTreeWidgetBuilder::updateChildren(const armem::Entity& entity, QTreeWidgetItem* entityItem)
    {
        snapshotBuilder.updateTree(entityItem, entity.history);
    }

    void MemoryTreeWidgetBuilder::updateChildren(const armem::EntitySnapshot& snapshot, QTreeWidgetItem* snapshotItem)
    {
        instanceBuilder.updateTree(snapshotItem, snapshot.instances);
    }

    void MemoryTreeWidgetBuilder::updateChildren(const armem::EntityData& data, QTreeWidgetItem* dataItem)
    {
        if (data.dataNavigator)
        {
            QTreeWidgetAronDataVisitor visitor(dataItem);
            visitor.applyTo(*data.dataNavigator);
        }
        else
        {
            dataItem->addChild(new QTreeWidgetItem(QStringList{"(No introspection available.)"}));
        }
    }

    QTreeWidgetItem* MemoryTreeWidgetBuilder::makeItem(const std::string& key, const armem::detail::MemoryLevel* level)
    {
        QStringList columns;
        if (level)
        {
            columns = QStringList
            {
                QString::fromStdString(level->getKeyString()),
                QString::fromStdString(simox::alg::capitalize_words(level->getLevelName())),
            };
        }
        else
        {
            columns = QStringList
            {
                QString::fromStdString(key),
                QString::fromStdString("(null)"),
            };
        }

        return new QTreeWidgetItem(columns);
    }

    QTreeWidgetItem* MemoryTreeWidgetBuilder::makeItem(const std::string& key, const armem::detail::MemoryContainer* container)
    {
        QTreeWidgetItem* item = makeItem(key, dynamic_cast<const detail::MemoryLevel*>(container));
        if (container)
        {
            item->setText(2, QString::number(container->size()));
        }
        return item;
    }

    void MemoryTreeWidgetBuilder::updateItem(const armem::detail::MemoryLevel* level, QTreeWidgetItem* item)
    {
        (void) level, (void) item;
    }

    void MemoryTreeWidgetBuilder::updateItem(
        const armem::detail::MemoryContainer* container, QTreeWidgetItem* item)
    {
        if (container)
        {
            item->setText(2, QString::number(container->size()));
        }
    }

}
