#include "CoreSegment.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error/ArMemError.h"


namespace armarx::armem
{

    CoreSegment::CoreSegment(const string& name, aron::type::AronTypePtr aronType) :
        Base(aronType)
    {
        this->name() = name;
    }

    CoreSegment::CoreSegment(const CoreSegment& other) : Base(other)
    {
        *this = other;
    }

    CoreSegment& CoreSegment::operator=(const CoreSegment& other)
    {
        other._copySelf(*this);
        other._copyElements(*this, true);
        return *this;
    }

    CoreSegment::~CoreSegment()
    {
    }

    const string& CoreSegment::name() const
    {
        return id().coreSegmentName;
    }

    void CoreSegment::_copyElements(CoreSegment& other, bool withData) const
    {
        other.providerSegments.clear();
        for (const auto& [name, providerSegment] : this->providerSegments)
        {
            other.addProviderSegment(providerSegment->copy(withData));
        }
    }


    bool CoreSegment::hasProviderSegment(const std::string& name) const
    {
        return providerSegments.count(name) > 0;
    }

    ProviderSegment& CoreSegment::getProviderSegment(const std::string& name)
    {
        return const_cast<ProviderSegment&>(const_cast<const CoreSegment*>(this)->getProviderSegment(name));
    }

    const ProviderSegment& CoreSegment::getProviderSegment(const std::string& name) const
    {
        auto it = this->providerSegments.find(name);
        if (it != providerSegments.end())
        {
            ARMARX_CHECK_NOT_NULL(it->second);
            return *it->second;
        }
        else
        {
            throw armem::error::MissingEntry("provider segment", name, getLevelName(), this->name());
        }
    }

    const Entity& CoreSegment::getEntity(const MemoryID& id) const
    {
        checkContainerName(id.coreSegmentName, this->name(), getLevelName());
        return getProviderSegment(id.providerSegmentName).getEntity(id);
    }


    ProviderSegment& CoreSegment::addProviderSegment(const std::string& name, aron::type::AronTypePtr providerSegmentType)
    {
        aron::type::AronTypePtr type = providerSegmentType ? providerSegmentType : this->aronType;
        return addProviderSegment(std::make_unique<ProviderSegment>(name, type));
    }

    ProviderSegment& CoreSegment::addProviderSegment(const ProviderSegment& providerSegment)
    {
        return addProviderSegment(std::make_unique<ProviderSegment>(providerSegment));
    }

    ProviderSegment& CoreSegment::addProviderSegment(ProviderSegmentPtr&& providerSegment)
    {
        if (hasProviderSegment(name()))
        {
            throw armem::error::ContainerEntryAlreadyExists("provider segment", name(), getLevelName(), this->name());
        }
        auto it = providerSegments.emplace(providerSegment->name(), std::move(providerSegment)).first;
        ARMARX_CHECK_NOT_NULL(it->second);
        return *it->second;
    }


    MemoryID CoreSegment::update(const InternalEntityUpdate& update)
    {
        checkContainerName(update.entityID.coreSegmentName, this->name(), getLevelName());

        auto it = this->providerSegments.find(update.entityID.providerSegmentName);
        if (it != providerSegments.end())
        {
            return it->second->update(update);
        }
        else
        {
            throw armem::error::MissingEntry("provider segment", update.entityID.providerSegmentName,
                                             getLevelName(), name());
        }
    }

}
