#include "Memory.h"

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error/ArMemError.h"


namespace armarx::armem
{

    Memory::Memory(const Memory& other) : EntityContainer(other)
    {
        *this = other;
    }

    Memory& Memory::operator=(const Memory& other)
    {
        other._copySelf(*this);
        other._copyElements(*this, true);
        return *this;
    }

    Memory::~Memory()
    {
    }


    const string& Memory::name() const
    {
        return id().memoryName;
    }

    void Memory::_copyElements(Memory& other, bool withData) const
    {
        other.coreSegments.clear();
        for (const auto& [name, coreSegment] : this->coreSegments)
        {
            other.coreSegments.emplace(name, coreSegment->copy(withData));
        }
    }


    bool Memory::hasCoreSegment(const std::string& name) const
    {
        return coreSegments.count(name) > 0;
    }

    CoreSegment& Memory::getCoreSegment(const std::string& name)
    {
        return const_cast<CoreSegment&>(const_cast<const Memory*>(this)->getCoreSegment(name));
    }
    const CoreSegment& Memory::getCoreSegment(const std::string& name) const
    {
        auto it = this->coreSegments.find(name);
        if (it != coreSegments.end())
        {
            ARMARX_CHECK_NOT_NULL(it->second);
            return *it->second;
        }
        else
        {
            throw armem::error::MissingEntry("core segment", name, getLevelName(), this->name());
        }
    }


    const Entity& Memory::getEntity(const MemoryID& id) const
    {
        checkContainerName(id.memoryName, this->name(), getLevelName());
        return getCoreSegment(id.coreSegmentName).getEntity(id);
    }

    CoreSegment& Memory::addCoreSegment(const string& name, aron::type::AronTypePtr coreSegmentType)
    {
        return addCoreSegment(std::make_unique<CoreSegment>(name, coreSegmentType));
    }

    CoreSegment& Memory::addCoreSegment(const CoreSegment& coreSegment)
    {
        return addCoreSegment(std::make_unique<CoreSegment>(coreSegment));
    }

    CoreSegment& Memory::addCoreSegment(CoreSegmentPtr&& coreSegment)
    {
        if (coreSegments.count(name()) > 0)
        {
            throw armem::error::ContainerEntryAlreadyExists("core segment", name(), getLevelName(), this->name());
        }
        auto it = coreSegments.emplace(coreSegment->name(), std::move(coreSegment)).first;
        ARMARX_CHECK_NOT_NULL(it->second);
        return *it->second;
    }

    std::vector<CoreSegment*> Memory::addCoreSegments(const std::vector<std::string>& names)
    {
        std::vector<CoreSegment*> segments;
        for (const auto& name : names)
        {
            try
            {
                segments.push_back(&addCoreSegment(name));
            }
            catch (const armem::error::ContainerEntryAlreadyExists& e)
            {
                ARMARX_INFO << e.what() << "\nIgnoring multiple addition.";
            }
        }
        return segments;
    }


    MemoryID Memory::update(const InternalEntityUpdate& update)
    {
        checkContainerName(update.entityID.memoryName, this->name(), getLevelName());

        auto it = this->coreSegments.find(update.entityID.coreSegmentName);
        if (it != coreSegments.end())
        {
            return it->second->update(update);
        }
        else
        {
            throw armem::error::MissingEntry("core segment", update.entityID.coreSegmentName,
                                             getLevelName(), this->name());
        }
    }


}
