#include "ProviderSegment.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error/ArMemError.h"


namespace armarx::armem
{

    ProviderSegment::ProviderSegment(const string& name, aron::type::AronTypePtr aronType) :
        Base(aronType)
    {
        this->name() = name;
    }

    ProviderSegment::ProviderSegment(const ProviderSegment& other) : Base(other)
    {
        *this = other;
    }

    ProviderSegment& ProviderSegment::operator=(const ProviderSegment& other)
    {
        other._copySelf(*this);
        other._copyElements(*this, true);
        return *this;
    }

    ProviderSegment::~ProviderSegment()
    {
    }

    const string& ProviderSegment::name() const
    {
        return id().providerSegmentName;
    }


    void ProviderSegment::_copySelf(ProviderSegment& other) const
    {
        Base::_copySelf(other);
        other.maxHistorySize = this->maxHistorySize;
    }

    void ProviderSegment::_copyElements(ProviderSegment& other, bool withData) const
    {
        other.entities.clear();
        for (const auto& [name, entity] : this->entities)
        {
            other.addEntity(entity->copy(withData));
        }
    }


    bool ProviderSegment::hasEntity(const string& name) const
    {
        return entities.count(name) > 0;
    }

    Entity& ProviderSegment::getEntity(const string& name)
    {
        return const_cast<Entity&>(const_cast<const ProviderSegment*>(this)->getEntity(name));
    }

    const Entity& ProviderSegment::getEntity(const std::string& name) const
    {
        auto it = this->entities.find(name);
        if (it != entities.end())
        {
            ARMARX_CHECK_NOT_NULL(it->second);
            return *it->second;
        }
        else
        {
            throw error::MissingEntry("entity", name, getLevelName(), this->name());
        }
    }

    const Entity& ProviderSegment::getEntity(const MemoryID& id) const
    {
        checkContainerName(id.providerSegmentName, this->name(), getLevelName());
        return getEntity(id.entityName);
    }

    MemoryID ProviderSegment::update(const InternalEntityUpdate& update)
    {
        checkContainerName(update.entityID.providerSegmentName, this->name(), getLevelName());

        auto it = this->entities.find(update.entityID.providerSegmentName);
        if (it == entities.end())
        {
            // Add entity entry.
            it = entities.emplace(update.entityID.entityName,
                                  std::make_unique<Entity>(update.entityID.entityName)).first;
            it->second->setMaxHistorySize(maxHistorySize);
        }
        // Update entity.
        return it->second->update(update);
    }

    Entity& ProviderSegment::addEntity(const string& name)
    {
        return addEntity(std::make_unique<Entity>(name));
    }

    Entity& ProviderSegment::addEntity(const Entity& entity)
    {
        return addEntity(std::make_unique<Entity>(entity));
    }

    Entity& ProviderSegment::addEntity(EntityPtr&& entity)
    {
        auto it = entities.emplace(entity->name(), std::move(entity)).first;
        return *it->second;
    }

    void ProviderSegment::setMaxHistorySize(long maxSize)
    {
        this->maxHistorySize = maxSize;
        for (auto& [name, entity] : entities)
        {
            entity->setMaxHistorySize(maxSize);
        }
    }

}
