#pragma once

#include <map>
#include <memory>
#include <string>

#include "Entity.h"
#include "detail/EntityContainer.h"


namespace armarx::armem
{

    class ProviderSegment;
    using ProviderSegmentPtr = std::unique_ptr<ProviderSegment>;

    /**
     * @brief Data of a provider segment containing multiple entities.
     */
    class ProviderSegment : public detail::TypedEntityContainer<EntityPtr, ProviderSegment>
    {
        using Base = detail::TypedEntityContainer<EntityPtr, ProviderSegment>;


    public:

        ProviderSegment(const std::string& name = "", aron::type::AronTypePtr aronType = nullptr) :
            Base(aronType)
        {
            this->name() = name;
        }

        ProviderSegment(const ProviderSegment& other);
        ProviderSegment& operator=(const ProviderSegment& other);

        ~ProviderSegment() override;


        using Base::name;
        const std::string& name() const override;


        bool hasEntity(const std::string& name) const;

        using Base::getEntity;
        const Entity& getEntity(const MemoryID& id) const override;

        Entity& getEntity(const std::string& name);
        const Entity& getEntity(const std::string& name) const;



        using Base::update;
        /**
         * @brief Updates an entity's history.
         *
         * Missing entity entries are added before updating.
         */
        virtual MemoryID update(const InternalEntityUpdate& update) override;

        /// Add an empty entity with the given name.
        Entity& addEntity(const std::string& name);
        /// Copy and insert an entity.
        Entity& addEntity(const Entity& entity);
        /// Move and insert an entity.
        Entity& addEntity(EntityPtr&& entity);


        /**
         * @brief Sets the maximum history size of entities in this segment.
         * This affects all current entities as well as new ones.
         * @see Entity::setMaxHistorySize()
         */
        void setMaxHistorySize(long maxSize);


    public:

        /// The entities.
        std::map<std::string, EntityPtr> entities;

        /**
         * @brief Maximum size of entity histories.
         * @see Entity::maxHstorySize
         */
        long maxHistorySize = -1;


        // MemoryLevel interface
    public:
        std::string getLevelName() const override
        {
            return "provider segment";
        }

        // MemoryContainer interface
    protected:
        ContainerT& _underlyingContainer() override
        {
            return entities;
        }

        void _copySelf(ProviderSegment& other) const override;
        void _copyElements(ProviderSegment& other, bool withData) const override;

    };

}
