#pragma once

#include <string>
#include <memory>
#include <type_traits>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "MemoryLevel.h"
#include "../../error.h"


namespace armarx::armem::detail
{

    /**
     * @brief A memory container at some point in the hierarchy.
     *
     * This class containes some convenience functions to allow more uniform
     * access to the different memory levels.
     */
    class MemoryContainer : public MemoryLevel
    {
    public:

        using MemoryLevel::MemoryLevel;
        virtual ~MemoryContainer();

        // Container methods

        virtual bool empty() const = 0;
        virtual std::size_t size() const = 0;
        virtual void clear() = 0;

    };


    /**
     * @class Provides default implmentations of `MemoryContainer`, as well as
     * iterators (which requires a template).
     */
    template <class _ContainerT, class Derived>
    class MemoryContainerImpl : public MemoryContainer
    {
    public:
        using ContainerT = _ContainerT;
        using Ptr = std::unique_ptr<Derived>;

    public:

        using MemoryContainer::MemoryContainer;
        virtual ~MemoryContainerImpl()
        {}


        // Container methods

        virtual bool empty() const
        {
            return _underlyingContainer().empty();
        }
        virtual std::size_t size() const
        {
            return _underlyingContainer().size();
        }
        virtual void clear()
        {
            return _underlyingContainer().clear();
        }

        auto begin() const
        {
            return _underlyingContainer().begin();
        }
        auto begin()
        {
            return _underlyingContainer().begin();
        }
        auto end() const
        {
            return _underlyingContainer().end();
        }
        auto end()
        {
            return _underlyingContainer().end();
        }


        // Copying

        virtual Ptr copy(bool withData = true) const
        {
            auto cast = dynamic_cast<const Derived*>(this);
            ARMARX_CHECK_NOT_NULL(cast);

            Ptr other = std::make_unique<Derived>();
            this->_copySelf(*other);
            this->_copyElements(*other, withData);
            return other;
        }

        /// Make a copy not containing any elements.
        virtual Ptr copyEmpty() const
        {
            auto cast = dynamic_cast<const Derived*>(this);
            ARMARX_CHECK_NOT_NULL(cast);

            Ptr other = std::make_unique<Derived>();
            this->_copySelf(*other);
            return other;
        }


    protected:


        virtual ContainerT& _underlyingContainer() = 0;
        virtual const ContainerT& _underlyingContainer() const
        {
            return const_cast<const ContainerT&>(const_cast<MemoryContainerImpl*>(this)->_underlyingContainer());
        }

        virtual void _copySelf(Derived& other) const
        {
            other.id() = this->id();
            (void) other;
        }
        virtual void _copyElements(Derived& other, bool withData) const
        {
            (void) other, (void) withData;
        }

    };

}
