#include "MemoryQueryProcessor.h"

#include <regex>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error/ArMemError.h"

#include "../memory/ice_conversions.h"


namespace armarx::armem
{

    void MemoryQueryProcessor::process(Memory& result, const query::MemoryQuery& query, const Memory& memory) const
    {
        if (auto q = dynamic_cast<const query::memory::All*>(&query))
        {
            process(result, *q, memory);
        }
        else if (auto q = dynamic_cast<const query::memory::Single*>(&query))
        {
            process(result, *q, memory);
        }
        else if (auto q = dynamic_cast<const query::memory::Regex*>(&query))
        {
            process(result, *q, memory);
        }
        else
        {
            throw armem::error::UnknownQueryType("memory segment", query);
        }
    }


    void MemoryQueryProcessor::process(Memory& result,
                                       const query::memory::All& query, const Memory& memory) const
    {
        for (const auto& [name, coreSegment] : memory.coreSegments)
        {
            result.addCoreSegment(coreSegmentProcessor.process(query.coreSegmentQueries, *coreSegment));
        }
    }

    void MemoryQueryProcessor::process(Memory& result,
                                       const query::memory::Single& query, const Memory& memory) const
    {
        try
        {
            const CoreSegment& coreSegment = memory.getCoreSegment(query.coreSegmentName);
            result.addCoreSegment(coreSegmentProcessor.process(query.coreSegmentQueries, coreSegment));
        }
        catch (const error::MissingEntry&)
        {
            // Leave empty.
        }
    }

    void MemoryQueryProcessor::process(Memory& result,
                                       const query::memory::Regex& query, const Memory& memory) const
    {
        std::regex regex(query.coreSegmentNameRegex);
        for (const auto& [name, coreSegment] : memory.coreSegments)
        {
            if (std::regex_search(coreSegment->name(), regex))
            {
                result.addCoreSegment(coreSegmentProcessor.process(query.coreSegmentQueries, *coreSegment));
            }
        }
    }


    data::MemoryPtr MemoryQueryProcessor::processToIce(
        const query::MemoryQuery& query, const Memory& memory) const
    {
        return toIce<data::MemoryPtr>(process(query, memory));
    }

    data::MemoryPtr MemoryQueryProcessor::processToIce(const query::MemoryQuerySeq& queries, const Memory& memory) const
    {
        return toIce<data::MemoryPtr>(process(queries, memory));
    }

}
