#include "ProviderSegmentQueryProcessor.h"

#include <regex>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error/ArMemError.h"

#include "../core/ice_conversions.h"


namespace armarx::armem
{

    void ProviderSegmentQueryProcessor::process(ProviderSegment& result, const query::ProviderSegmentQuery& query, const ProviderSegment& providerSegment) const
    {
        if (auto q = dynamic_cast<const query::provider::All*>(&query))
        {
            process(result, *q, providerSegment);
        }
        else if (auto q = dynamic_cast<const query::provider::Single*>(&query))
        {
            process(result, *q, providerSegment);
        }
        else if (auto q = dynamic_cast<const query::provider::Regex*>(&query))
        {
            process(result, *q, providerSegment);
        }
        else
        {
            throw armem::error::UnknownQueryType("provider segment", query);
        }
    }


    void ProviderSegmentQueryProcessor::process(ProviderSegment& result,
            const query::provider::All& query, const ProviderSegment& providerSegment) const
    {
        for (const auto& [name, entity] : providerSegment.entities)
        {
            result.addEntity(std::move(*entityProcessor.process(query.entityQueries, *entity)));
        }
    }

    void ProviderSegmentQueryProcessor::process(ProviderSegment& result,
            const query::provider::Single& query, const ProviderSegment& providerSegment) const
    {
        try
        {
            const Entity& entity = providerSegment.getEntity(query.entityName);
            result.addEntity(std::move(*entityProcessor.process(query.entityQueries, entity)));
        }
        catch (const error::MissingEntry&)
        {
            // Leave empty.
        }
    }

    void ProviderSegmentQueryProcessor::process(ProviderSegment& result,
            const query::provider::Regex& query, const ProviderSegment& providerSegment) const
    {
        std::regex regex(query.entityNameRegex);
        for (const auto& [name, entity] : providerSegment.entities)
        {
            if (std::regex_search(entity->name(), regex))
            {
                result.addEntity(std::move(*entityProcessor.process(query.entityQueries, *entity)));
            }
        }
    }


    data::ProviderSegmentPtr ProviderSegmentQueryProcessor::processToIce(
        const query::ProviderSegmentQuery& query, const ProviderSegment& providerSegment) const
    {
        return toIce<data::ProviderSegmentPtr>(process(query, providerSegment));
    }

}
