/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "AronPCLPointcloudTypeNavigator.h"

namespace armarx
{
    namespace aron
    {
        namespace typenavigator
        {
            // constructors
            AronPCLPointcloudTypeNavigator::AronPCLPointcloudTypeNavigator(const AronPath& path) :
                navigator::AronNavigator<AronTypeDescriptor, type::AronType>::AronNavigator(AronTypeDescriptor::eAronPCLPointcloudType, path),
                AronTypeNavigator(AronTypeDescriptor::eAronPCLPointcloudType, path),
                AronNDArraySerializerTypeNavigator(AronTypeDescriptor::eAronEigenMatrixType, path),
                type(new type::AronPCLPointcloudType({0, 0}, ""))
            {
            }

            AronPCLPointcloudTypeNavigator::AronPCLPointcloudTypeNavigator(const type::AronPCLPointcloudTypePtr& o, const AronPath& path) :
                navigator::AronNavigator<AronTypeDescriptor, type::AronType>::AronNavigator(AronTypeDescriptor::eAronPCLPointcloudType, path),
                AronTypeNavigator(AronTypeDescriptor::eAronPCLPointcloudType, path),
                AronNDArraySerializerTypeNavigator(AronTypeDescriptor::eAronEigenMatrixType, path),
                type(o)
            {
                CheckAronPtrForNull("AronPCLPointcloudTypeNavigator", "AronPCLPointcloudTypeNavigator", getPath(), o);
            }

            std::vector<int> AronPCLPointcloudTypeNavigator::getDimensions() const
            {
                return type->dimensions;
            }

            void AronPCLPointcloudTypeNavigator::setDimensions(const std::vector<int>& d)
            {
                if (d.size() != 2)
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointcloudTypeNavigator", "setDimensions", "Wrong number of elements in dimension array. The size must be equal to 2. The size was: " + std::to_string(d.size()), getPath());
                }
                type->dimensions = d;
            }

            unsigned int AronPCLPointcloudTypeNavigator::getWidth() const
            {
                return type->dimensions[0];
            }

            unsigned int AronPCLPointcloudTypeNavigator::getHeight() const
            {
                return type->dimensions[1];
            }

            std::string AronPCLPointcloudTypeNavigator::getUsedType() const
            {
                return type->typeName;
            }

            void AronPCLPointcloudTypeNavigator::setWidth(const unsigned int& w)
            {
                if (w == 0)
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointcloudTypeNavigator", "setWidth", "The width of aa pointcloud cannot be 0", getPath());
                }
                type->dimensions[0] = w;
            }

            void AronPCLPointcloudTypeNavigator::setHeight(const unsigned int& h)
            {
                if (h == 0)
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointcloudTypeNavigator", "setHeight", "The height of aa pointcloud cannot be 0", getPath());
                }
                type->dimensions[1] = h;
            }

            void AronPCLPointcloudTypeNavigator::setUsedType(const std::string& u)
            {
                if (u.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointcloudTypeNavigator", "setUsedType", "The useType of a pointcloud cannot be empty", getPath());
                }
                type->typeName = u;
            }

            // static methods
            AronPCLPointcloudTypeNavigatorPtr AronPCLPointcloudTypeNavigator::DynamicCast(const AronTypeNavigatorPtr& n)
            {
                CheckTypeNavigatorPtrForNull("AronPCLPointcloudTypeNavigator", "DynamicCast[Before]", n);
                AronPCLPointcloudTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronPCLPointcloudTypeNavigator>(n);
                CheckTypeNavigatorPtrForNull("AronPCLPointcloudTypeNavigator", "DynamicCast[After]", casted);
                return casted;
            }

            // virtual implementations
            type::AronTypePtr AronPCLPointcloudTypeNavigator::getResult() const
            {
                if (type->dimensions.size() != 2)
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointcloudTypeNavigator", "getResult", "The dimension size is wrong. Got size: " + std::to_string(type->dimensions.size()), getPath());
                }
                if (std::any_of(type->dimensions.begin(), type->dimensions.end(), [](int i)
            {
                return i <= 0;
            }))
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointcloudTypeNavigator", "getResult", "The dimension size is wrong. At least one empty is <=0", getPath());
                }
                if (type->typeName.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointcloudTypeNavigator", "getResult", "The useType of image is empty.", getPath());
                }
                return type;
            }
        }
    }
}

