#pragma once

#include "../datanavigator/AronAllDataNavigators.h"


namespace armarx::aron::visitor
{

    /**
     * @brief A visitor traversing an `aron::data::AronDataPtr` using
     * `aron::datanavigator::AronDataNavigator`s, taking care of type checking,
     * casting and hierarchical traversal.
     *
     * To use it, create a class deriving from `AronDataVisitor` and
     * override the overloads of `visitEnter()`, `visitExit()` and `visit()`
     * for the types you would like to handle (see blow for a copy-and-paste
     * example).
     * Then, instantiate your derived visitor and call `applyTo()`, passing
     * the `AronDataPtr` or the root `AronDataNavigator`.
     *
     * For dicts and lists, you will get a `visitEnter()`/`visitExit()` pair.
     * For types without children (e.g. primitive types), you get a single
     * `visit()` call.
     *
     * Each visit function should return true to continue traversal or
     * false to stop it. All default implementations return true.
     *
     * To the the location of the passed navigator in the original `AronData`,
     * use `navigator.getPath()`.
     *
     * @see  `aron::data::AronDataPtr`
     * @see  `aron::datanavigator::AronDataNavigator`
     * @see `std::stack` To manage data with stack semantics.
     */
    class AronDataVisitor
    {
    public:

        // Define types to ease writing code in this and derived classes.

        using AronDataNavigator = datanavigator::AronDataNavigator;

        using AronDictDataNavigator = datanavigator::AronDictDataNavigator;
        using AronListDataNavigator = datanavigator::AronListDataNavigator;

        using AronBoolDataNavigator = datanavigator::AronBoolDataNavigator;
        using AronDoubleDataNavigator = datanavigator::AronDoubleDataNavigator;
        using AronFloatDataNavigator = datanavigator::AronFloatDataNavigator;
        using AronIntDataNavigator = datanavigator::AronIntDataNavigator;
        using AronLongDataNavigator = datanavigator::AronLongDataNavigator;
        using AronNDArrayDataNavigator = datanavigator::AronNDArrayDataNavigator;
        using AronStringDataNavigator = datanavigator::AronStringDataNavigator;


    public:

        AronDataVisitor();
        virtual ~AronDataVisitor();


        bool applyTo(data::AronDataPtr& aronData);
        bool applyTo(AronDataNavigator& navigator);



        virtual bool visitEnter(AronDictDataNavigator& dict)
        {
            (void) dict;
            return true;
        }
        virtual bool visitExit(AronDictDataNavigator& dict)
        {
            (void) dict;
            return true;
        }
        virtual bool visitEnter(AronListDataNavigator& list)
        {
            (void) list;
            return true;
        }
        virtual bool visitExit(AronListDataNavigator& list)
        {
            (void) list;
            return true;
        }

        virtual bool visit(AronBoolDataNavigator& b)
        {
            (void) b;
            return true;
        }
        virtual bool visit(AronDoubleDataNavigator& d)
        {
            (void) d;
            return true;
        }
        virtual bool visit(AronFloatDataNavigator& f)
        {
            (void) f;
            return true;
        }
        virtual bool visit(AronIntDataNavigator& i)
        {
            (void) i;
            return true;
        }
        virtual bool visit(AronLongDataNavigator& l)
        {
            (void) l;
            return true;
        }
        virtual bool visit(AronNDArrayDataNavigator& array)
        {
            (void) array;
            return true;
        }
        virtual bool visit(AronStringDataNavigator& string)
        {
            (void) string;
            return true;
        }


    private:

        bool applyToChildren(AronDataNavigator& navigator);

    };

}


// Copy-and-paste example below. Remove functions you don't need.
#if 0

struct MyDerivedAronDataVisitor : public armarx::aron::visitor::AronDataVisitor
{

    bool visitEnter(AronDictDataNavigator& dict) override
    {
        (void) dict;
        return true;
    }
    bool visitExit(AronDictDataNavigator& dict) override
    {
        (void) dict;
        return true;
    }
    bool visitEnter(AronListDataNavigator& list) override
    {
        (void) list;
        return true;
    }
    bool visitExit(AronListDataNavigator& list) override
    {
        (void) list;
        return true;
    }

    bool visit(AronBoolDataNavigator& b) override
    {
        (void) b;
        return true;
    }
    bool visit(AronDoubleDataNavigator& d) override
    {
        (void) d;
        return true;
    }
    bool visit(AronFloatDataNavigator& f) override
    {
        (void) f;
        return true;
    }
    bool visit(AronIntDataNavigator& i) override
    {
        (void) i;
        return true;
    }
    bool visit(AronLongDataNavigator& l) override
    {
        (void) l;
        return true;
    }
    bool visit(AronNDArrayDataNavigator& array) override
    {
        (void) array;
        return true;
    }
    bool visit(AronStringDataNavigator& string) override
    {
        (void) string;
        return true;
    }

};

#endif
